function PLLHexData = setfreq(fVCO, ICName, fREF)
% PLLHexData = setfreq(fVCO, ICName, fREF) Determines the four 24-bit
% Hexidecimal codes to sequentially write to Analog Devices ADF4XXX
% PLLs ICNAME to cause a voltage-controlled oscillator (VCO) to output FVCO
% provided a reference frequency FREF.
% 
% All frequencies are in MHz. If not specified, FREF is assumed
% to be the standard RF reference frequency of 10 MHz.
%
% This bit of code works for the following Analog Devices PLLs (ICNAME):
%   ADF4001 -> 20 MHz to 200 MHz
%   ADF4111 -> 80 MHz to 1200 GHz
%   ADF4107 -> 1 GHz to 7 GHz
%
% It will probably work for the most PLLs in the ADF4XXX family
% by adding the chip details in this files switch statement.

%==========================================================================
% Big switch statement describing PLLs.
% ADD NEW PLLs HERE AND GIVE PARAMETERS (see datasheet)
%==========================================================================
switch ICName
    case 'ADF4107'
        Plist=[8 16 32 64]; % dual-modulus prescalar values
        hasAB = 1; % has an A and B register
        minRF = 1000; maxRF = 7000; % min and max RF input freq
        maxPrescalarOutFreq = 300; % max allowable prescalar output freq
    case 'ADF4001'
        Plist= 1; % no dual-modulus prescalar
        hasAB = 0; % no "A" and "B" registers... instead has an N register
        minRF = 20; maxRF = 200; % min and max RF input freq
        maxPrescalarOutFreq = 200; % no prescalar so same as maxRF
    case 'ADF4111'
        Plist=[8 16 32 64]; % dual-modulus prescalar values
        hasAB = 1; % has an A and B register
        minRF = 80; maxRF = 1200; % min and max RF input freq
        maxPrescalarOutFreq = 165; % max allowable prescalar output freq
    otherwise
        display('Unknown PLL.  Add to list?');
        return;
end
%==========================================================================
if~exist('fREF','var') | isempty(fREF)
    fREF = 10; % Assume reference frequency is 10 MHz
end

if (fVCO < minRF || fVCO > maxRF) % ...Not sure why we limit to 6 GHz...
    disp(['Invalid frequency for ', ICName,'.']);
    return;
end

% Default registers to 0
funcReg =false(1,24);
RReg = false(1,24);
ABReg = false(1,24);

jjStart = find(fVCO./Plist<=maxPrescalarOutFreq,1); % determines the first prescalar value to try
if isempty(jjStart)
    disp(['Requested output frequency (',num2str(fVCO),' MHz) is too large.']);
    return;
end

ratio = fVCO/fREF;
Rlist = 1:1000; % If R (reference frequency divider) is too big, PLL has a hard time locking.
P = []; R = [];
flag = 1; ii = 1;
% Search for smallest R that will still allow desired frequency.
while flag & ii<=length(Rlist)
    jj = jjStart;
    while jj<=length(Plist) & flag
        % determine range
        if hasAB % we have a 13-bit A register and 6-bit B register :)
             maxVal = Plist(jj)*8191 + 63;
             minVal = Plist(jj)*3;
        else % we only have a 13-bit N register :(
            maxVal = Plist(jj)*8191;
            minVal = Plist(jj)*1;
        end
        % Determine if we have a match
        if (ratio*Rlist(ii) >= minVal) && (ratio*Rlist(ii) <= maxVal) && (mod(ratio*Rlist(ii),1)<=1e-6)
             R = Rlist(ii);
             P = Plist(jj);
             flag = 0;
        else
            jj = jj+1;
        end
    end
    ii = ii+1;
end

if isempty(P) | isempty(R)
    disp('No R value found'); % Failed to find a match
    disp(['ratio = ',num2str(ratio),', min = ',num2str(minVal),', max = ',num2str(maxVal)]);
    return;
end

% set prescalar bits
switch P
    case 1
        funcReg(1+0) = 0;
        funcReg(1+1) = 0;
    case 8
        funcReg(1+0) = 0;
        funcReg(1+1) = 0;
    case 16
        funcReg(1+0) = 0;
        funcReg(1+1) = 1;
    case 32
        funcReg(1+0) = 1;
        funcReg(1+1)= 0;
    case 64
        funcReg(1+0) = 1;
        funcReg(1+1) = 1;
end

% set divide-by-R bits
N = R;
ii = 21;
while (N ~= 0) & ii~=7
      RReg(1+ii) = mod(N,2);
      N = floor(N/2);
      ii=ii-1;
end

% set bits for B (or N) 
if hasAB % B is B
    B = floor((fVCO*R)/(fREF*P));
else % B is actually N 
    B = floor(fVCO*R/fREF);
end
N = B;
ii = 15;
while N ~=0 & ii ~=2
	ABReg(1+ii) = mod(N,2);
    N = floor(N/2);
    ii=ii-1;
end
    
% determine A (if hasAB=0, then effectively A==0)
A = floor((fVCO*R)/fREF - P*B);
N = A;
ii = 21;
while N~=0 & ii~=15
    ABReg(1+ii) = mod(N,2);
      N = floor(N/2);
      ii=ii-1;
end
  
% sanity check to make sure the calculated frequency equals the desired
% frequency
if (~((fVCO == (P*B+A)*fREF/R) || (fVCO == B*fREF/R)))
    disp('ERROR: calculated fVCO does not match specified fVCO');
    disp(['fVCO = ',num2str(fVCO),' MHz']);
    disp(['fCalc = ',num2str((P*B+A)*fREF/R),' MHz']);
    disp(['P = ',num2str(P)]);
    disp(['R = ',num2str(R)]);
    disp(['A = ',num2str(A)]);
    disp(['B = ',num2str(B)]);
    return;
else
    % spit out values for the curious
    disp(['P = ',num2str(P)]);
    disp(['R = ',num2str(R)]);
    disp(['A = ',num2str(A)]);
    disp(['B = ',num2str(B)]);
end

% Fill in Defaults

% Max Current Settings
funcReg(1+3) = 0;
funcReg(1+4) = 0;
funcReg(1+5) = 0;
funcReg(1+6) = funcReg(1+3);
funcReg(1+7) = funcReg(1+4);
funcReg(1+8) = funcReg(1+5);

% Counter Reset
funcReg(1+21) = 1;

% Lock Detect
RReg(1+3) = 0;

% Anti Backlash Pulse Width
RReg(1+6) = 0;
RReg(1+7) = 1;

% Control Bits
funcReg(1+22) = 1;
ABReg(1+23) = 1;

% MUX Out
funcReg(1+17) = 0;
funcReg(1+18) = 1;
funcReg(1+19) = 0;

% Phase Detection Polarity
funcReg(1+16) = 1;

% Converts the binary register information to hexidecimal:
% initialize the "function" register
PLLHexData(1,:) = printBinaryToHex(funcReg);
% Provide the "R" register data
PLLHexData(2,:)= printBinaryToHex(RReg);
% Provide the "AB" (or "N") register data
PLLHexData(3,:)= printBinaryToHex(ABReg);
% Change a bit in the function register (tells the PLL to start phase-lockin') 
funcReg(1+21) = 0;
PLLHexData(4,:)= printBinaryToHex(funcReg);

% Prints the PLL write data
disp(PLLHexData);

%==========================================================================
function data = printBinaryToHex(dataStream)
% converts binary as hex (... this is probably a built in MATLAB function,
% but oh well...)
data =[];
for ii=0:(length(dataStream)/4-1)
    num = 0;
    if dataStream(1+ii*4)
        num=num+8;
    end
    if dataStream(1+ii*4+1)
        num=num+4;
    end
    if dataStream(1+ii*4+2)
        num=num+2;
    end
    if dataStream(1+ii*4+3)
        num=num+1;
    end
    data = [data, sprintf('%01X',num)];
end